"""
Plot SUVI L2 Coronal Hole Boundaries
====================================
Purpose:
  Use Python to plot the SUVI L2 coronal hole boundary (chbnd) product in 3 coordinate systems
  (Heliographic Stonyhurst, Heliographic Carrington, and Pixel).
"""


# %%
# Overview
# ------------

# %%
# In this example, we will be plotting the coronal hole boundaries (chbnd_loc) from
# the coronal hole boundary product. The boundaries are determined by outlining the coronal hole classifications from
# the SUVI Thematic Map algorithm.
#
# We will plot chbnd_loc for the following 3 coordinate systems (R-Theta is currently unavailable):
#  * Heliographic Stonyhurst `(longitude, latitude)`
#  * Heliographic Carrington `(longitude, latitude)`
#  * Pixel `(x pixels, y pixels)`

# %%
# Imports
# -----------

# %%
# First, we will import the necessary libraries:
__authors__ = "elucas, ajarvis"

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
from datetime import datetime, timedelta
import netCDF4 as nc
from astropy.coordinates import SkyCoord
import sunpy
import sunpy.map
import astropy.units as u
from sunpy.coordinates import frames
from sunpy.net import Fido, attrs as a
from astropy.io import fits


# %%
# Helper Functions
# -----------------------------------------
# We need to define a few helper functions for use throughout plotting:

# Fetch the SUVI SunPy map to use as a base for plotting.
def create_suvi_sunpymap(date, goes=16, wavelength=131, rng=2):
    ds0 = (date - timedelta(minutes=rng)).strftime("%Y/%m/%d %H:%M:%S")
    ds1 = (date + timedelta(minutes=rng)).strftime("%Y/%m/%d %H:%M:%S")
    q = Fido.search(a.Time(ds0, ds1), a.Instrument.suvi, a.Wavelength(wavelength * u.angstrom))
    tmp_files = Fido.fetch(q)
    # Select files for level 2 composites only
    for tmp_file in tmp_files:
        if (f'g{goes}' in tmp_file) and ('l2' in tmp_file):
            data, header = fits.getdata(tmp_file), fits.getheader(tmp_file, 1)
            suvi_map = sunpy.map.Map(data, header)
            return suvi_map, data, header
    print('No SUVI map available')
    return None

# %%


# Convert time given in the .nc file to datetime.
def convert_time(time_nc):
    date_2000 = datetime(2000, 1, 1, 12, 0)
    date = date_2000 + timedelta(seconds=time_nc)
    return date

# %%


# Define legend for this product.
def legend_handles():
    markers = ['o']
    fillstyle = ['full']
    labels = ['chbnd_loc']
    f = lambda m, fill: plt.plot([], [], marker=m, color='k', ls='-', fillstyle=fill)[0]
    handles = [f(markers[i], fillstyle[i]) for i in range(len(markers))]
    return handles, labels

# %%
# Retrieving the SUVI SunPy Map
# -------------------------------

# %%
#| Let's use some of the helper functions above to retrieve a SUVI image.
#| We'll grab an image of the sun in the 195Å wavelength for datetime 2022-11-27 20:24.
#| Example data file: `dr_suvi-l2-chbnd_g16_s20221127T202400Z_e20221127T202800Z_v1-0-5.nc <../../_static/suvi/dr_suvi-l2-chbnd_g16_s20221127T202400Z_e20221127T202800Z_v1-0-5.nc>`__

date = datetime(2022, 11, 27, 20, 24)
goes = 16
wavelength = 195
suvi_chbnd_path = f'../../data/suvi/dr_suvi-l2-chbnd_g16_s20221127T202400Z_e20221127T202800Z_v1-0-5.nc'
chbnd_nc = nc.Dataset(suvi_chbnd_path)

# %%
# Let's look at the variables contained in the coronal hole boundary files.
for k in chbnd_nc.variables.keys():
    print(k)

# %%
# For our examples, we will be looking at the chbnd_loc variable.
# We can see the general structure by inspecting this variable in any coordinate system, for example,
# in Heliographic Stonyhurst.

print(chbnd_nc['chbnd_loc_hg'])

# %%
#| The structure in this example is `(1, 2, 16, 2)`:
#|  `1` : The first dimension of the structure is the time dimension, which will always have 1 value for this product.
#|  `2` : The second dimension is the number of separate coronal holes within this image.
#|  `16`: The third dimension is each vertex of the coronal hole boundary polygon. We are limited to 16 vertices for legacy operational reasons.
#|  `2` : The fourth dimension is the coordinate pair of each vertex, with units as described above.

# %%
# We need to extract the time (found in the 'time' variable), and convert to python datetime in order to fetch a
# corresponding SUVI SunPy map at the closest time possible.

# Get file time
chbnd_time = convert_time(chbnd_nc['time'][:][0])
# Access SUVI SunPy map for that time and defined wavelength
suvi_map, suvi_data, suvi_header = create_suvi_sunpymap(chbnd_time, goes=goes, wavelength=wavelength)

# %%
# Plotting the Flare Locations
# ----------------------------
# Now that the SUVI SunPy map has been retrieved, we can overplot the coordinates from the .nc file onto that image.
#
# `Note: Heliographic Stonyhurst and Heliographic Carrington are only defined on-disk, so any features
# (or pieces of features) that extend off-disk will not be plotted.`

# %%
# Heliographic Stonyhurst Coordinates
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

# Plot the SUVI SunPy composite image
fig_hg = plt.figure(figsize=(10, 10))
fig_hg.tight_layout(h_pad=-50)
fig_hg.suptitle(f'Heliographic Stonyhurst Coordinates, GOES-{goes}')
ax_hg = fig_hg.add_subplot(projection=suvi_map)
suvi_map.plot(axes=ax_hg, clip_interval=(0.1, 99.9) * u.percent)
num_points = 100

# Extract coronal hole boundary array from netCDF
chbnd_hg = chbnd_nc['chbnd_loc_hg'][:].data[0]
colors = cm.Reds_r(np.linspace(0.3, 0.6, len(chbnd_hg)))

for hole, c in zip(chbnd_hg, colors):
    hole[hole == -9999.] = np.nan
    # Close boundary polygon
    hole = [*hole, hole[0]]

    chbnd_hgs_sc = SkyCoord([(float(v[0]), float(v[1])) * u.degree for v in hole], obstime=date,
                            observer=suvi_map.observer_coordinate, frame=frames.HeliographicStonyhurst)
    # Transform SkyCoord object to the SUVI SunPy map's coordinate frame
    chbnd_hgs_sc = chbnd_hgs_sc.transform_to(suvi_map.coordinate_frame)

    # Plot boundary
    ax_hg.plot_coord(chbnd_hgs_sc, color=c, marker='o', markersize=2)

# Add legend for features
handles, labels = legend_handles()
plt.legend(handles, labels, loc=1, framealpha=1)

plt.show()

# %%
# Heliographic Carrington Coordinates
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

# Plot the SUVI SunPy composite image
fig_car = plt.figure(figsize=(10, 10))
fig_car.tight_layout(h_pad=-50)
fig_car.suptitle(f'Heliographic Carrington Coordinates, GOES-{goes}')
ax_car = fig_car.add_subplot(projection=suvi_map)
suvi_map.plot(axes=ax_car, clip_interval=(0.1, 99.9) * u.percent)
num_points = 100

# Extract coronal hole boundary array from netCDF
chbnd_car = chbnd_nc['chbnd_loc_car'][:].data[0]
colors = cm.YlOrBr_r(np.linspace(0.3, 0.6, len(chbnd_car)))

for hole, c in zip(chbnd_car, colors):
    hole[hole == -9999.] = np.nan
    # Close boundary polygon
    hole = [*hole, hole[0]]

    chbnd_car_sc = SkyCoord([(float(v[0]), float(v[1])) * u.degree for v in hole], obstime=date,
                            observer=suvi_map.observer_coordinate, frame=frames.HeliographicCarrington)
    # Transform SkyCoord object to the SUVI SunPy map's coordinate frame
    chbnd_car_sc = chbnd_car_sc.transform_to(suvi_map.coordinate_frame)

    # Plot boundary
    ax_car.plot_coord(chbnd_car_sc, color=c, marker='o', markersize=2)

# Add legend for features
handles, labels = legend_handles()
plt.legend(handles, labels, loc=1, framealpha=1)

plt.show()

# %%
# Pixel Coordinates
# ^^^^^^^^^^^^^^^^^

# Plot the SUVI SunPy composite image
fig_px = plt.figure(figsize=(10, 10))
fig_px.tight_layout()
ax_px = fig_px.add_subplot(projection=suvi_map)
suvi_map.plot(axes=ax_px, clip_interval=(0.1, 99.9) * u.percent)
fig_px.suptitle(f'Pixel Coordinates, GOES-{goes}')
num_points = 100

# Extract coronal hole boundary array from netCDF
chbnd_pix = chbnd_nc['chbnd_loc_pix'][:].data[0]
colors = cm.Greens_r(np.linspace(0.3, 0.6, len(chbnd_pix)))

for hole, c in zip(chbnd_pix, colors):
    hole[hole == -9999.] = np.nan
    # Close boundary polygon
    hole = [*hole, hole[0]]

    chbnd_pix_x = [h[0] for h in hole]
    chbnd_pix_y = [h[1] for h in hole]

    # Plot boundary
    ax_px.plot_coord(chbnd_pix_x, chbnd_pix_y, color=c, marker='o', markersize=2)

# Add legend for features
handles, labels = legend_handles()
plt.legend(handles, labels, loc=1, framealpha=1)

plt.show()
